# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""RoiAlign function for the MultiScaleRoIAlign class"""

from typing import List, Union

import torch
from torch import Tensor
from torch.nn.modules.utils import _pair
from torch.jit.annotations import BroadcastingList2
from torchvision.extension import _assert_has_ops


def roi_align(
    inp: Tensor,
    boxes: Union[Tensor, List[Tensor]],
    output_size: BroadcastingList2[int],
    spatial_scale: float = 1.0,
    sampling_ratio: int = -1,
    aligned: bool = False,
) -> Tensor:
    # pylint: disable=too-many-arguments
    """
    Performs Region of Interest (RoI) Align operator with average pooling, as described in
    Mask R-CNN.

    Args:
        inp (Tensor[N, C, H, W]): The input tensor, i.e. a batch with ``N`` elements. Each
            element contains ``C`` feature maps of dimensions ``H x W``.
            If the tensor is quantized, we expect a batch size of ``N == 1``.
        boxes (Tensor[K, 5] or List[Tensor[L, 4]]): the box coordinates in (x1, y1, x2, y2)
            format where the regions will be taken from.
            The coordinate must satisfy ``0 <= x1 < x2`` and ``0 <= y1 < y2``.
            If a single Tensor is passed, then the first column should
            contain the index of the corresponding element in the batch, i.e. a number in
            ``[0, N - 1]``.
            If a list of Tensors is passed, then each Tensor will correspond to the boxes for an
            element i in the batch.
        output_size (int or Tuple[int, int]): the size of the output (in bins or pixels) after
            the pooling is performed, as (height, width).
        spatial_scale (float): a scaling factor that maps the input coordinates to
            the box coordinates. Default: 1.0
        sampling_ratio (int): number of sampling points in the interpolation grid
            used to compute the output value of each pooled output bin. If > 0,
            then exactly ``sampling_ratio x sampling_ratio`` sampling points per bin are used. If
            <= 0, then an adaptive number of grid points are used (computed as
            ``ceil(roi_width / output_width)``, and likewise for height). Default: -1
        aligned (bool): If False, use the legacy implementation.
            If True, pixel shift the box coordinates it by -0.5 for a better alignment with the two
            neighboring pixel indices. This version is used in Detectron2

    Returns:
        Tensor[K, C, output_size[0], output_size[1]]: The pooled RoIs.
    """
    _assert_has_ops()
    check_roi_boxes_shape(boxes)
    rois = boxes
    output_size = _pair(output_size)
    if not isinstance(rois, Tensor):
        rois = convert_boxes_to_roi_format(rois)
    return torch.ops.torchvision.roi_align(
        inp,
        rois,
        spatial_scale,
        output_size[0],
        output_size[1],
        sampling_ratio,
        aligned,
    )


def _cat(tensors: List[Tensor], dim: int = 0) -> Tensor:
    """
    Efficient version of torch.cat that avoids a copy if there is only a single element in a list
    """
    assert isinstance(tensors, (list, tuple))
    if len(tensors) == 1:
        return tensors[0]
    return torch.cat(tensors, dim)


def convert_boxes_to_roi_format(boxes: List[Tensor]) -> Tensor:
    """Converts RoI boxes from List to Tensor format, and includes the respective box indices
    that corresponds to the element i in a batch

    Args:
        boxes (List[Tensor[L, 4]]): RoI boxes with the format of (x1, y1, x2, y2)

    Returns:
        Tensor: The converted RoI boxes
    """
    concat_boxes = _cat(boxes, dim=0)
    temp = []
    for i, box in enumerate(boxes):
        temp.append(torch.full_like(box[:, :1], i))
    ids = _cat(temp, dim=0)
    rois = torch.cat([ids, concat_boxes], dim=1)
    return rois


def check_roi_boxes_shape(boxes: Union[Tensor, List[Tensor]]) -> None:
    """Checks the RoI boxes shape, and raising error if the shape is not correct.

    If the 'boxes' is an instance of Tensor, the first column should contain the index of the
    corresponding element in the batch, i.e. a number in ``[0, N - 1]``. Hence, it should be in the
    shape Tensor[K, 5].

    Otherwise, if the 'boxes' is an instance of List, each Tensor in the list should correspond to
    the boxes for an element i in the batch. Hence, it should be in the shape List[Tensor[K, 4]],
    and the axis with dimension of 4 corresponds to the box coordinates in (x1, y1, x2, y2).

    Output type is depending on the format of proposals generated by the RPN.

    Args:
        boxes (Union[Tensor, List[Tensor]]): RoI boxes
    """
    if isinstance(boxes, (list, tuple)):
        for _tensor in boxes:
            assert (
                _tensor.size(1) == 4
            ), "The shape of the tensor in the boxes list is not correct as List[Tensor[L, 4]]"
    elif isinstance(boxes, Tensor):
        assert (
            boxes.size(1) == 5
        ), "The boxes tensor shape is not correct as Tensor[K, 5]"
    else:
        assert False, "boxes is expected to be a Tensor[L, 5] or a List[Tensor[K, 4]]"



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Core PeekingDuck CLI commands."""

import logging
import os
import tempfile
from pathlib import Path
from time import perf_counter
from typing import Any, Dict, List, Union

import click
import yaml

from peekingduck.commands import LOGGER_NAME
from peekingduck.runner import Runner
from peekingduck.utils.deprecation import deprecate
from peekingduck.utils.logger import LoggerSetup
from peekingduck.viewer import Viewer

logger = logging.getLogger(LOGGER_NAME)  # pylint: disable=invalid-name


@click.command()
@click.option("--custom_folder_name", default="custom_nodes")
def init(custom_folder_name: str) -> None:
    """Initializes a PeekingDuck project."""
    print("Welcome to PeekingDuck!")
    _create_custom_folder(custom_folder_name)
    _create_pipeline_config_yml()


@click.command()
@click.option(
    "--config_path",
    default=None,
    type=click.Path(),
    help=(
        "List of nodes to run. None assumes pipeline_config.yml at current working directory"
    ),
)
@click.option(
    "--log_level",
    default="info",
    help="""Modify log level {"critical", "error", "warning", "info", "debug"}""",
)
@click.option(
    "--node_config",
    default="None",
    help="""Modify node configs by wrapping desired configs in a JSON string.\n
        Example: --node_config '{"node_name": {"param_1": var_1}}'""",
)
@click.option(
    "--num_iter",
    default=None,
    type=int,
    help="Stop pipeline after running this number of iterations",
)
@click.option(
    "--viewer",
    default=False,
    is_flag=True,
    help="Launch PeekingDuck viewer",
)
def run(  # pylint: disable=too-many-arguments
    config_path: str,
    log_level: str,
    node_config: str,
    num_iter: int,
    viewer: bool,
    nodes_parent_dir: str = "src",
) -> None:
    """Runs PeekingDuck"""
    LoggerSetup.set_log_level(log_level)

    if config_path is None:
        curr_dir = Path.cwd()
        if (curr_dir / "pipeline_config.yml").is_file():
            config_path = curr_dir / "pipeline_config.yml"
        elif (curr_dir / "run_config.yml").is_file():
            deprecate(
                "using 'run_config.yml' as the default pipeline configuration "
                "file is deprecated and will be removed in the future. Please "
                "use 'pipeline_config.yml' instead.",
                2,
            )
            config_path = curr_dir / "run_config.yml"
        else:
            config_path = curr_dir / "pipeline_config.yml"
    pipeline_config_path = Path(config_path)

    if viewer:
        logger.info("Launching PeekingDuck Viewer")
        start_time = perf_counter()
        pkd_viewer = Viewer(
            pipeline_path=pipeline_config_path,
            config_updates_cli=node_config,
            custom_nodes_parent_subdir=nodes_parent_dir,
            num_iter=num_iter,
        )
        end_time = perf_counter()
        logger.debug(f"Startup time = {end_time - start_time:.2f} sec")
        pkd_viewer.run()
    else:
        start_time = perf_counter()
        runner = Runner(
            pipeline_path=pipeline_config_path,
            config_updates_cli=node_config,
            custom_nodes_parent_subdir=nodes_parent_dir,
            num_iter=num_iter,
        )
        end_time = perf_counter()
        logger.debug(f"Startup time = {end_time - start_time:.2f} sec")
        runner.run()


@click.command()
def verify_install() -> None:
    """Verifies PeekingDuck installation by running object detection on
    'wave.mp4'.
    """
    LoggerSetup.set_log_level("info")

    cwd = Path.cwd()
    with tempfile.TemporaryDirectory() as tmp_dir:
        pipeline_config_path = Path(tmp_dir) / "verification_pipeline.yml"

        _create_pipeline_config_yml(
            [
                {
                    "input.visual": {
                        "source": "https://storage.googleapis.com/peekingduck/videos/wave.mp4"
                    }
                },
                "model.yolo",
                "draw.bbox",
                "output.screen",
            ],
            pipeline_config_path,
        )

        runner = Runner(
            pipeline_path=pipeline_config_path,
            config_updates_cli="None",
            custom_nodes_parent_subdir="src",
            num_iter=None,
        )
        runner.run()
        os.chdir(cwd)


def _create_custom_folder(custom_folder_name: str) -> None:
    """Makes custom nodes folder to create custom nodes.

    Args:
        custom_folder_name (:obj:`str`): Name of the custom nodes folder.
    """
    curr_dir = Path.cwd()
    custom_nodes_dir = curr_dir / "src" / custom_folder_name
    custom_nodes_config_dir = custom_nodes_dir / "configs"

    logger.info(f"Creating custom nodes folder in {custom_nodes_dir}")
    custom_nodes_dir.mkdir(parents=True, exist_ok=True)
    custom_nodes_config_dir.mkdir(parents=True, exist_ok=True)


def _create_pipeline_config_yml(
    default_nodes: List[Union[str, Dict[str, Any]]] = None,
    default_path: Path = Path("pipeline_config.yml"),
) -> None:
    """Initializes the declarative *pipeline_config.yml*.

    Args:
        default_nodes (List[Union[str, Dict[str, Any]]]): A list of PeekingDuck
            node. For nodes with custom configuration, use a dictionary instead
            of a string.
        default_path (Path): Path of the pipeline config file.
    """
    # Default yml to be discussed
    if default_nodes is None:
        default_nodes = [
            {
                "input.visual": {
                    "source": "https://storage.googleapis.com/peekingduck/videos/wave.mp4"
                }
            },
            "model.posenet",
            "draw.poses",
            "output.screen",
        ]
    default_yml = dict(nodes=default_nodes)

    with open(default_path, "w") as yml_file:
        yaml.dump(default_yml, yml_file, default_flow_style=False)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Transformation class for Generalized RCNN class.
Modifications include:
- Removed if conditions for keypoints detection
- Removed training / target related code and parameters
- Removed tracing related codes
- Removed torch_choice method (unused)
"""

from typing import Dict, Iterable, List, Optional, Tuple, Union
import math
import torch
from torch import nn, Tensor
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detection.roi_heads import (
    paste_masks_in_image,
)
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detection.image_list import (
    ImageList,
)


class GeneralizedRCNNTransform(nn.Module):
    """
    Performs input transformation before feeding the data to a GeneralizedRCNN
    model.

    The transformations it perform are:
        - input normalization (mean subtraction and std division)
        - input resizing to match min_size / max_size

    It returns a ImageList for the inputs
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        min_size: int,
        max_size: int,
        image_mean: Iterable[float],
        image_std: Iterable[float],
        size_divisible: int = 32,
        fixed_size: Optional[Tuple[int, int]] = None,
    ):
        super().__init__()
        if not isinstance(min_size, (list, tuple)):
            min_size = (min_size,)  # type: ignore[assignment]
        self.min_size = min_size
        self.max_size = max_size
        self.image_mean = image_mean
        self.image_std = image_std
        self.size_divisible = size_divisible
        self.fixed_size = fixed_size

    def forward(self, images: List[Tensor]) -> ImageList:
        """Normalizes and resizes the images, follow by padding the images to the same size and
        putting the images into the ImageList object that contains the padded images and the
        original image sizes

        Args:
            images (List[Tensor]): Input images

        Raises:
            ValueError: Raise error if the image does not have the shape [C, H, W]

        Returns:
            ImageList: An object containing padded images and the original image sizes
        """
        for i, image in enumerate(images):

            if image.dim() != 3:
                raise ValueError(
                    "images is expected to be a list of 3d tensors "
                    f"of shape [C, H, W], got {image.shape}"
                )
            image = self.normalize(image)
            image = self.resize(image)
            images[i] = image

        image_sizes = [img.shape[-2:] for img in images]
        padded_images = self.batch_images(images, size_divisible=self.size_divisible)
        image_sizes_list: List[Tuple[int, int]] = []
        for image_size in image_sizes:
            assert len(image_size) == 2
            image_sizes_list.append((image_size[0], image_size[1]))

        image_list = ImageList(padded_images, image_sizes_list)
        return image_list

    def normalize(self, image: Tensor) -> Tensor:
        """Normalize image"""
        if not image.is_floating_point():
            raise TypeError(
                "Expected input images to be of floating type (in range [0, 1]), "
                f"but found type {image.dtype} instead"
            )
        dtype, device = image.dtype, image.device
        mean = torch.as_tensor(self.image_mean, dtype=dtype, device=device)
        std = torch.as_tensor(self.image_std, dtype=dtype, device=device)
        return (image - mean[:, None, None]) / std[:, None, None]

    def resize(self, image: Tensor) -> Tensor:
        """Resize image"""
        # Note: min_size is either a list or tuple
        size = float(self.min_size[-1])  # type: ignore[index]
        image = _resize_image_and_masks(
            image, size, float(self.max_size), self.fixed_size
        )

        return image

    def batch_images(self, images: List[Tensor], size_divisible: int = 32) -> Tensor:
        """Pad the images in a batch to the same size"""
        max_size = self.max_by_axis([list(img.shape) for img in images])
        stride = float(size_divisible)
        max_size = list(max_size)
        max_size[1] = int(math.ceil(float(max_size[1]) / stride) * stride)
        max_size[2] = int(math.ceil(float(max_size[2]) / stride) * stride)

        batch_shape = [len(images)] + max_size
        batched_imgs = images[0].new_full(batch_shape, 0)
        for i in range(batched_imgs.shape[0]):
            img = images[i]
            batched_imgs[i, : img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)

        return batched_imgs

    @staticmethod
    def max_by_axis(the_list: List[List[int]]) -> List[int]:
        """Find the maximum size for each axis in a batch of images"""
        maxes = the_list[0]
        for sublist in the_list[1:]:
            for index, item in enumerate(sublist):
                maxes[index] = max(maxes[index], item)
        return maxes

    @staticmethod
    def postprocess(
        result: List[Dict[str, Tensor]],
        image_shapes: List[Tuple[int, int]],
        original_image_sizes: List[Tuple[int, int]],
    ) -> List[Dict[str, Tensor]]:
        """Postprocess the result"""
        for i, (pred, im_s, o_im_s) in enumerate(
            zip(result, image_shapes, original_image_sizes)
        ):
            boxes = pred["boxes"]
            boxes = resize_boxes(boxes, im_s, o_im_s)
            result[i]["boxes"] = boxes
            if "masks" in pred:
                masks = pred["masks"]
                masks = paste_masks_in_image(masks, boxes, o_im_s)
                result[i]["masks"] = masks

        return result

    def __repr__(self) -> str:
        format_string = self.__class__.__name__ + "("
        indent = "\n    "
        format_string += (
            f"{indent}Normalize(mean={self.image_mean}, std={self.image_std})"
        )
        format_string += (
            f"{indent}Resize(min_size={self.min_size}, "
            f"max_size={self.max_size}, mode='bilinear')"
        )
        format_string += "\n)"
        return format_string


def resize_boxes(
    boxes: Tensor,
    original_size: Union[List[int], Tuple[int, int]],
    new_size: Union[List[int], Tuple[int, int]],
) -> Tensor:
    """Resize bounding boxes"""
    ratios = [
        torch.tensor(s, dtype=torch.float32, device=boxes.device)
        / torch.tensor(s_orig, dtype=torch.float32, device=boxes.device)
        for s, s_orig in zip(new_size, original_size)
    ]
    ratio_height, ratio_width = ratios
    xmin, ymin, xmax, ymax = boxes.unbind(1)

    xmin = xmin * ratio_width
    xmax = xmax * ratio_width
    ymin = ymin * ratio_height
    ymax = ymax * ratio_height
    return torch.stack((xmin, ymin, xmax, ymax), dim=1)


def _resize_image_and_masks(
    image: Tensor,
    self_min_size: float,
    self_max_size: float,
    fixed_size: Optional[Tuple[int, int]] = None,
) -> Tensor:
    """Resize image and masks"""
    im_shape = torch.tensor(image.shape[-2:])

    size: Optional[List[int]] = None
    scale_factor: Optional[float] = None
    recompute_scale_factor: Optional[bool] = None
    if fixed_size is not None:
        size = [fixed_size[1], fixed_size[0]]
    else:
        min_size = torch.min(im_shape).to(dtype=torch.float32)
        max_size = torch.max(im_shape).to(dtype=torch.float32)
        scale = torch.min(self_min_size / min_size, self_max_size / max_size)

        scale_factor = scale.item()
        recompute_scale_factor = True

    image = nn.functional.interpolate(  # type: ignore[call-arg]
        image[None],
        size=size,
        scale_factor=scale_factor,
        mode="bilinear",
        recompute_scale_factor=recompute_scale_factor,
        align_corners=False,
    )[0]

    return image



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements the PlayList single-column list view class"""

from typing import Callable, Dict, Union
import logging
import tkinter as tk
from tkinter import ttk
from peekingduck.viewer.playlist import PipelineStats, PlayList

OP_LIST = ["add", "delete", "run"]  # Supported GUI operations
PLAYLIST_WIDTH = 200


class SingleColumnPlayListView:  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """Use tk.ListBox as single-column list view"""

    def __init__(self, playlist: PlayList, root: tk.Widget):
        self.logger = logging.getLogger(__name__)
        self.playlist = playlist
        self.root = root
        self.view = None
        self._callback: Dict[str, Callable] = {}
        self._selected: Union[None, str] = None
        self._sort_desc = False
        self._tk_fg_system = ""  # different for diff platforms
        self.create_tk_widgets()
        self.redraw_view()

    def create_tk_widgets(self) -> None:
        """Create Tk widgets for the GUI view"""
        left_margin_frm = ttk.Frame(
            self.root, name="left_margin_frm", width=50, height=100
        )
        left_margin_frm.pack(side=tk.LEFT, fill=tk.NONE, expand=False)

        lbl = tk.Label(self.root, text="Pipelines:")
        lbl.pack(side=tk.TOP)
        lbl.bind("<Button-1>", self.header_press)
        self.header = lbl

        # playlist controls
        playlist_ctrl_frm = ttk.Frame(master=self.root, name="playlist_ctrl")
        playlist_ctrl_frm.pack(side=tk.BOTTOM)

        self._btn_add = ttk.Button(playlist_ctrl_frm, text="Add")
        self._btn_delete = ttk.Button(playlist_ctrl_frm, text="Delete")
        self._btn_play = ttk.Button(playlist_ctrl_frm, text="Run")
        # Tk technotes:
        # - if bind to Button (mouse down) event,
        #   the Play button will be "stuck" and show button down style (blue on macOS)
        #   when it is clicked on.
        # - binding to ButtonRelease (mouse up) event solves above issue
        self._btn_add.bind("<ButtonRelease-1>", self.btn_add_press)
        self._btn_delete.bind("<ButtonRelease-1>", self.btn_delete_press)
        self._btn_play.bind("<ButtonRelease-1>", self.btn_run_press)
        for child in playlist_ctrl_frm.winfo_children():
            child.pack(side=tk.LEFT)  # pack above buttons

        # info panel
        info_frm = ttk.Frame(master=self.root, name="playlist_info")
        info_frm.pack(side=tk.BOTTOM, fill=tk.X)

        lbl = tk.Label(info_frm, text="Pipeline Information:")
        # info labels
        lbl.grid(row=0, column=0, columnspan=2)
        lbl = tk.Label(info_frm, text="Name:", anchor=tk.E)
        lbl.grid(row=1, column=0, sticky="ne")
        lbl = tk.Label(info_frm, text="Modified:", anchor=tk.E)
        lbl.grid(row=2, column=0, sticky="ne")
        lbl = tk.Label(info_frm, text="Path:", anchor=tk.E)
        lbl.grid(row=3, column=0, sticky="ne")
        # playlist info
        self._info_name = tk.Message(
            info_frm, text="name", width=PLAYLIST_WIDTH, anchor=tk.W
        )
        self._info_name.grid(row=1, column=1, sticky="nw")
        self._info_datetime = tk.Message(
            info_frm, text="datetime", width=PLAYLIST_WIDTH, anchor=tk.W
        )
        self._info_datetime.grid(row=2, column=1, sticky="nw")
        self._info_path = tk.Message(
            info_frm, text="path", width=PLAYLIST_WIDTH, anchor=tk.W
        )
        self._info_path.grid(row=3, column=1, sticky="nw")

        num_col, _ = info_frm.grid_size()  # config column sizes
        for i in range(num_col):
            info_frm.grid_columnconfigure(i, weight=1)

        # listbox
        playlist_listbox = tk.Listbox(
            master=self.root, relief=tk.RIDGE, borderwidth=1, height=20
        )
        playlist_listbox.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        playlist_listbox.bind("<<ListboxSelect>>", self.selection_changed)
        self.tk_listbox = playlist_listbox

    def redraw_view(self) -> None:
        """Populate playlist contents"""
        self.sorted_playlist = sorted(self.playlist, reverse=self._sort_desc)
        self._pipeline_to_index_map: Dict[str, int] = {}
        self._index_to_stats_map: Dict[int, PipelineStats] = {}
        self.tk_listbox.delete(0, tk.END)
        for i, stats in enumerate(self.sorted_playlist):
            self._pipeline_to_index_map[stats.pipeline] = i
            self._index_to_stats_map[i] = stats
            self.tk_listbox.insert(i, stats.name)
            if len(self._tk_fg_system) == 0:
                # save platform specific foreground color
                self._tk_fg_system = self.tk_listbox["fg"]
                self.logger.debug(f"saving system foreground='{self._tk_fg_system}'")
            if len(stats.datetime) == 0:
                self.tk_listbox.itemconfig(i, {"fg": "red"})  # mark as error
        self.header["text"] = f"Pipelines: {'v' if self._sort_desc else '^'}"

    def get_selected_index(self) -> int:
        """Return index of selected listbox entry

        Returns:
            int: Index of selected entry
        """
        selection_indices = self.tk_listbox.curselection()
        i = selection_indices[0]  # only want first selection
        return i

    def register_callback(self, operation: str, player_callback: Callable) -> None:
        """Register callback function in Player to be called when playlist events are generated

        Args:
            operation (str): One of [ "add", "delete", "run" ]
            player_callback (Callable): The hook to the Player
        """
        if operation in OP_LIST:
            self._callback[operation] = player_callback
        else:
            raise ValueError(f"Unsupported callback operation: {operation}")

    def reset(self) -> None:
        """Force widget to recalculate width based on contents"""
        self.tk_listbox.config(width=0)

    def select(self, pipeline: str) -> None:
        """Select given pipeline on the GUI

        Args:
            pipeline (str): Pipeline to select.
        """
        self._selected = pipeline
        self.show_selected_pipeline()

    def show_selected_pipeline(self) -> None:
        """Show current selected pipeline by highlighting it in the listbox"""
        if self._selected:
            self.tk_listbox.selection_clear(0, tk.END)
            i = self._pipeline_to_index_map[self._selected]
            self.tk_listbox.select_set(i)
            self.update_pipeline_info(i)

    def update_pipeline_info(self, i: int) -> None:
        """Update pipeline info panel details"""
        stats = self._index_to_stats_map[i]
        self.logger.debug(f"show stats[{i}]: {stats.name}")
        self._info_name["text"] = stats.name
        if stats.datetime:
            display_datetime = f"{stats.datetime}"
            self._info_datetime["foreground"] = self._tk_fg_system
        else:
            display_datetime = "missing file"
            self._info_datetime["foreground"] = "red"
        self._info_datetime["text"] = display_datetime
        self._info_path["text"] = stats.pipeline

    #
    # Tk event callbacks
    #
    def btn_add_press(self, event: tk.Event) -> None:  # pylint: disable=unused-argument
        """Callback to handle "+" button

        Args:
            event (tk.Event): Tk event object
        """
        self.logger.debug("btn_add_press")
        if self._callback["add"]():
            self.redraw_view()

    def btn_delete_press(
        self, event: tk.Event  # pylint: disable=unused-argument
    ) -> None:
        """Callback to handle "-" button

        Args:
            event (tk.Event): Tk event object
        """
        i = self.get_selected_index()
        stats = self._index_to_stats_map[i]
        self.logger.debug(f"btn_delete_press: {stats.pipeline}")
        if self._callback["delete"](pipeline=stats.pipeline):
            self.redraw_view()

    def btn_run_press(self, event: tk.Event) -> None:  # pylint: disable=unused-argument
        """Callback to handle "Run" button

        Args:
            event (tk.Event): Tk event object
        """
        i = self.get_selected_index()
        stats = self._index_to_stats_map[i]
        self.logger.debug(f"btn_run_press: {stats.pipeline}")
        self._callback["run"](pipeline=stats.pipeline)

    def header_press(self, event: tk.Event) -> None:  # pylint: disable=unused-argument
        """Callback when header is clicked, changes sort order

        Args:
            event (tk.Event): Tk event object.
        """
        self._sort_desc = not self._sort_desc
        self.redraw_view()
        self.show_selected_pipeline()

    def selection_changed(
        self, event: tk.Event  # pylint: disable=unused-argument
    ) -> None:
        """Update pipeline info when user selects a pipeline in playlist

        Args:
            event (tk.Event): Tk event object
        """
        i = self.get_selected_index()
        self.update_pipeline_info(i)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implement PeekingDuck Viewer
"""

from typing import List
from contextlib import redirect_stderr
from pathlib import Path
import logging
from io import StringIO
import os
import platform
import traceback
import tkinter as tk
from tkinter import filedialog
from tkinter.messagebox import askyesno, showerror
import threading
import copy
import cv2
import numpy as np
from PIL import Image, ImageTk
from peekingduck.declarative_loader import DeclarativeLoader
from peekingduck.pipeline.pipeline import Pipeline
from peekingduck.viewer.playlist import PlayList
from peekingduck.viewer.viewer_gui import create_window
from peekingduck.viewer.viewer_utils import (
    get_keyboard_char,
    get_keyboard_modifier,
)

####################
# Globals
####################
BUTTON_DELAY: int = 250  # milliseconds (0.25 of a second)
BUTTON_REPEAT: int = int(1000 / 60)  # milliseconds (60 fps)
STOP_PIPELINE_DELAY: float = 2.0  # seconds before switching pipelines
FPS_60: int = int(1000 / 60)  # milliseconds per iteration
ZOOM_TEXT: List[str] = ["50%", "75%", "100%", "125%", "150%", "200%", "250%", "300%"]
ZOOM_DEFAULT_IDX: int = 2
ZOOMS: List[float] = [0.5, 0.75, 1.0, 1.25, 1.50, 2.00, 2.50, 3.00]  # > 3x is slow!
PLAY_BUTTON_TEXT = "Play"
STOP_BUTTON_TEXT = "Stop"


def parse_streams(io_stream: StringIO) -> str:
    """Helper method to parse I/O streams.
    Used to capture errors/exceptions from PeekingDuck.

    Args:
        io_stream (StringIO): the I/O stream to parse

    Returns:
        str: parsed stream
    """
    msg = io_stream.getvalue()
    msg = os.linesep.join([s for s in msg.splitlines() if s])
    return msg


class Viewer:  # pylint: disable=too-many-instance-attributes, too-many-public-methods
    """Implement PeekingDuck Viewer class"""

    def __init__(
        self,
        pipeline_path: Path,
        config_updates_cli: str,
        custom_nodes_parent_subdir: str,
        num_iter: int = 0,
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.config_updates_cli = config_updates_cli
        self.custom_nodes_parent_path = custom_nodes_parent_subdir
        self.num_iter = num_iter
        # init PlayList object
        self.home_path = Path.home()
        self.playlist = PlayList(self.home_path)
        self.playlist.load_playlist_file()
        self.playlist_show: bool = False
        # init pipeline to run
        self.init_pipeline(pipeline_path)
        self.playlist.add_pipeline(self.pipeline_full_path)
        # configure keyboard shortcuts map
        self._keyboard_shortcuts = {
            "=": self._zoom_reset,
            "+": self._zoom_in,
            "-": self._zoom_out,
        }
        # forward type hinting of internal working vars
        self._frames: List[np.ndarray] = []
        self._frame_idx: int = -1
        self.zoom_idx: int = ZOOM_DEFAULT_IDX
        self.is_output_playback: bool = False
        self.is_pipeline_running: bool = False
        self.state: str = "play"

    def run(self) -> None:
        """Main method to setup Viewer and run Tk event loop"""
        self.logger.debug(f"cwd={Path.cwd()}")
        self.logger.debug(f"pipeline={self.pipeline_path}")
        # create Tkinter window and frames
        create_window(self)
        # trap macOS Cmd-Q keystroke
        if platform.system() == "Darwin":
            self.logger.debug("binding macOS cmd-Q")
            self.root.createcommand("::tk::mac::Quit", self.on_exit)
        # activate internal timer function and start Tkinter event loop
        self.timer_function()
        self.root.mainloop()

    ####################
    #
    # Tk Event Handlers
    #
    ####################
    def btn_hide_show_playlist_press(self) -> None:
        """Handle Hide/Show Playlist button

        Tk technotes:
            - Behavior:
                Playlist on right is fixed width.  When image is expanded, it should not
                cover playlist.  But when playlist is hidden and revealed, the expanding
                image will cover playlist
            - To fix above, need to
              a) also pack_forget() the image frame, along with the playlist frame
              b) when revealing playlist, pack() playlist first, then pack() image
              c) now the expanding image will not cover the playlist
        """
        if self.playlist_show:
            self.tk_playlist_frm.pack_forget()
        else:
            self.tk_playlist_frm.pack(side=tk.RIGHT, fill=tk.Y)
            self.tk_playlist_view.reset()
            self.tk_playlist_view.select(str(self.pipeline_full_path))
        self.playlist_show = not self.playlist_show

    def btn_play_stop_press(self) -> None:
        """Handle Play/Stop button"""
        self.logger.debug(f"btn_play_stop_press start: self.state={self.state}")
        if self.is_pipeline_running:
            self.stop_running_pipeline()
        elif self.is_output_playback:
            self.stop_playback()
        else:
            self.start_playback()
        self.logger.debug(f"btn_play_stop_press end: self.state={self.state}")

    def btn_zoom_in_press(self) -> None:
        """Zoom in: make image larger"""
        self._zoom_in()

    def btn_zoom_out_press(self) -> None:
        """Zoom out: make image smaller"""
        self._zoom_out()

    def on_exit(self) -> None:
        """Handle viewer quit event"""
        self.logger.info("quitting viewer")

        if self.is_pipeline_running or self.is_output_playback:
            self.logger.debug("stopping current pipeline")
            if self.is_pipeline_running:
                self.run_pipeline_end()
            elif self.is_output_playback:
                self.stop_playback()

            # add non-blocking wait to let background task clean up properly
            self.logger.debug(f"wait {STOP_PIPELINE_DELAY} sec")
            wait_event = threading.Event()
            wait_event.wait(STOP_PIPELINE_DELAY)

        self.cancel_timer_function()
        self.logger.debug("saving playlist")
        self.playlist.save_playlist_file()
        self.root.destroy()

    def on_keypress(self, event: tk.Event) -> None:
        """Handle all keydown events.
        Default system shortcuts are automatically handled, e.g. CMD-Q quits on macOS

        Args:
            event (tk.Event): the key down event
        """
        self.logger.debug(
            f"keypressed: char={event.char}, keysym={event.keysym}, state={event.state}"
        )
        key_state: int = int(event.state)
        mod = get_keyboard_modifier(key_state)
        key = get_keyboard_char(event.char, event.keysym)
        self.logger.debug(f"mod={mod}, key={key}")
        # handle supported keyboard shortcuts here
        if mod.startswith("ctrl"):
            if key in self._keyboard_shortcuts:
                self._keyboard_shortcuts[key]()

    def on_resize(self, event: tk.Event) -> None:
        """Handle window resize event.

        Args:
            event (tk.Event): The resize event.
        """
        if str(event.widget) == ".":
            # NB: "." is the root widget, i.e. main window
            self.logger.debug(
                f"on_resize: widget={event.widget}, h={event.height}, w={event.width}"
            )

    def on_add_pipeline(self) -> bool:
        """Add pipeline to playlist

        Returns:
            bool: True if pipeline added, False otherwise
        """
        # filetypes = (("Pipeline files", "*.yml"), ("All files", "*.*"))
        filetypes = [("Pipeline files", "*.yml")]
        pipeline_filepath = filedialog.askopenfilename(
            title="Open a pipeline file (*.yml)",
            initialdir=self.home_path,
            filetypes=filetypes,
        )
        self.logger.debug(f"on add pipeline: filepath={pipeline_filepath}")
        if pipeline_filepath:
            self.playlist.add_pipeline(pipeline_filepath)
            return True
        return False

    def on_delete_pipeline(self, pipeline: str) -> bool:
        """Delete pipeline from playlist

        Args:
            pipeline (str): Pipeline to delete
        """
        self.logger.debug(f"on delete pipeline {pipeline}")
        answer = askyesno(
            title="Confirm Delete Pipeline from Playlist",
            message=f"Are you sure you want to delete\n\n{pipeline}\n\nfrom playlist?",
        )
        if answer:
            self.playlist.delete_pipeline(pipeline)
            return True
        return False

    def on_run_pipeline(self, pipeline: str) -> None:
        """Callback function for Run pipeline

        Args:
            pipeline (str): Pipeline to execute
        """
        # run new pipeline
        self.logger.debug(f"on run pipeline {pipeline}")
        if self.is_pipeline_running:
            self.run_pipeline_end()
        elif self.is_output_playback:
            self.stop_playback()

        # add non-blocking wait to let background task clean up properly
        self.logger.debug(f"wait {STOP_PIPELINE_DELAY} sec")
        wait_event = threading.Event()
        wait_event.wait(STOP_PIPELINE_DELAY)

        # double check status variables
        self.logger.debug(f"self.state={self.state}")
        self.logger.debug(f"self.is_output_playback={self.is_output_playback}")
        self.logger.debug(f"self.is_pipeline_running={self.is_pipeline_running}")
        if self.state == "play" or self.is_output_playback or self.is_pipeline_running:
            self.logger.warning(
                "Inconsistent internal self states:\n"
                f"  state={self.state} - expect !play"
                f"  is_output_playback={self.is_output_playback} - expect False"
                f"  is_pipeline_running={self.is_pipeline_running} - expect False"
            )

        # start new pipeline here
        self.cancel_timer_function()
        self.init_pipeline(Path(pipeline))
        self.timer_function()

    ##################
    #
    # Background Timer
    #
    ##################
    def timer_function(self) -> None:
        """Function to do background processing in Tkinter's way"""
        if self.state == "play":
            # Only two states: 1) playing back video or 2) executing pipeline
            if self.is_output_playback:
                self.do_playback()
            else:
                # Executing pipeline: check which execution state we are in
                if not self.is_pipeline_running:
                    self.run_pipeline_start()
                elif self._pipeline.terminate:
                    self.run_pipeline_end()
                else:
                    self.run_pipeline_one_iteration()

        self.root.update()  # wake up GUI
        self._bkgd_job = self.tk_header.after(FPS_60, self.timer_function)

    def cancel_timer_function(self) -> None:
        """Cancel the background timer function"""
        if self._bkgd_job:
            self.tk_header.after_cancel(self._bkgd_job)
            self._bkgd_job = None

    ####################
    #
    # Display Management
    #
    ####################
    def _backward_one_frame(self) -> bool:
        """Move back one frame, can be called repeatedly"""
        if self._frame_idx > 0:
            self._frame_idx -= 1
            self._update_slider_and_show_frame()
            return True
        return False

    def _forward_one_frame(self) -> bool:
        """Move forward one frame, can be called repeatedly"""
        if self._frame_idx + 1 < len(self._frames):
            self._frame_idx += 1
            self._update_slider_and_show_frame()
            return True
        return False

    def _show_frame(self) -> None:
        """Display image frame pointed to by frame_idx"""
        if self._frames:
            frame = self._frames[self._frame_idx]
            frame = self._apply_zoom(frame)
            # self.logger.debug(f"show_frame {self.frame_idx} size={frame.shape}")
            img_arr = Image.fromarray(frame)
            img_tk = ImageTk.PhotoImage(img_arr)
            # self.logger.debug(f"img_tk: {img_tk.width()}x{img_tk.height()}")
            self._img_tk = img_tk  # save to avoid python GC
            self.tk_output_image.config(image=img_tk)

    def _apply_zoom(self, frame: np.ndarray) -> np.ndarray:
        """Zoom output image according to current zoom setting

        Args:
            frame (np.ndarray): image frame data to be zoomed

        Returns:
            np.ndarray: the zoomed image
        """
        if self.zoom_idx != ZOOM_DEFAULT_IDX:
            zoom = ZOOMS[self.zoom_idx]
            new_size = (
                int(frame.shape[0] * zoom),
                int(frame.shape[1] * zoom),
                frame.shape[2],
            )
            frame = cv2.resize(frame, (new_size[1], new_size[0]))
        return frame

    def _zoom_in(self) -> None:
        """Zoom in on image"""
        if self.zoom_idx + 1 < len(ZOOMS):
            self.zoom_idx += 1
            self._update_zoom_and_show_frame()

    def _zoom_out(self) -> None:
        """Zoom out on image"""
        if self.zoom_idx > 0:
            self.zoom_idx -= 1
            self._update_zoom_and_show_frame()

    def _zoom_reset(self) -> None:
        """Reset zoom to default 1x"""
        self.zoom_idx = ZOOM_DEFAULT_IDX
        self._update_zoom_and_show_frame()

    def _update_zoom_and_show_frame(self) -> None:
        """Update zoom widget and refresh current frame"""
        glyph = ZOOM_TEXT[self.zoom_idx]
        self.tk_lbl_zoom["text"] = f"{glyph}"
        self._show_frame()

    def _set_header_playing(self) -> None:
        """Change header text to playing..."""
        self.tk_header["text"] = f"Playing {self.pipeline_path.name}"

    def _set_header_running(self) -> None:
        """Change header text to running..."""
        self.tk_header["text"] = f"Running {self.pipeline_path.name}"

    def _set_header_stop(self) -> None:
        """Change header text to pipeline pathname"""
        self.tk_header["text"] = f"{self.pipeline_path.name}"

    def _update_slider_and_show_frame(self) -> None:
        """Update slider based on frame index and show new frame"""
        frame_num = self._frame_idx + 1
        self.tk_scale.set(frame_num)
        self.tk_lbl_frame_num["text"] = frame_num
        self._show_frame()

    def _enable_progress(self) -> None:
        """Show progress bar and hide slider"""
        self.logger.debug("enable progress")
        self.tk_scale.grid_remove()  # hide slider
        self.tk_progress.grid()  # show progress bar

    def _enable_slider(self) -> None:
        """Show slider and hide progress bar"""
        self.logger.debug("enable slider")
        self.tk_progress.grid_remove()  # hide progress bar
        self.tk_scale.grid()  # show slider
        self.tk_scale.configure(from_=1, to=len(self._frames))
        self.sync_slider_to_frame(self.tk_scale.get())

    def _set_status_text(self, text: str) -> None:
        """Set the status bar text to given text string.

        Args:
            text (str): the status bar text
        """
        self.tk_status_bar["text"] = text

    def sync_slider_to_frame(self, val: str) -> None:
        """Update frame index based on slider value change

        Args:
            val (str): slider value
        """
        self.logger.debug(f"sync slider to frame: {val} {type(val)}")
        idx = int(self.tk_scale.get())
        idx_start = self.tk_scale["from"]
        idx_end = self.tk_scale["to"]
        self.logger.debug(f"idx={idx}, from:{idx_start}, to:{idx_end})")
        if idx >= len(self._frames):
            idx = len(self._frames)
        self._frame_idx = idx - 1
        self.tk_lbl_frame_num["text"] = idx
        self._show_frame()

    ####################
    #
    # Pipeline Execution
    #
    ####################
    def init_pipeline(self, pipeline_path: Path) -> None:
        """Initialise pipeline to be run

        Args:
            pipeline_path (Path): The pipeline to run
        """
        self.pipeline_path = pipeline_path
        is_abs = pipeline_path.is_absolute()
        self.logger.debug(
            f"init pipeline: pipeline_path={pipeline_path}, is_abs={is_abs}"
        )
        # expand pipeline path if required
        if not is_abs:
            full_path = pipeline_path.resolve()
            self.logger.debug(f"full path: {full_path}")
            self.pipeline_full_path = full_path
        else:
            self.pipeline_full_path = pipeline_path
        # set custom nodes path accordingly
        self.custom_nodes_parent_path = str(self.pipeline_full_path.parent / "src")
        self.logger.debug(f"custom nodes parent: {type(self.custom_nodes_parent_path)}")
        # init internal working vars
        self._frames = []
        self._frame_idx = -1
        self.zoom_idx = ZOOM_DEFAULT_IDX
        self.is_output_playback = False
        self.is_pipeline_running = False
        self.state = "play"  # activate auto play (cf. self.timer_function)
        self.bkgd_job = None

    def pipeline_error(self, exc_msg: str, err_stream: StringIO) -> None:
        """Helper method to handle pipeline error conditions

        Args:
            exc_msg (str): message from exception object
            err_stream (StringIO): error I/O stream
        """
        self.logger.error("Error when running pipeline:")
        self.logger.error(f"Exception msg: {exc_msg}")
        err_msg = parse_streams(err_stream)
        self.logger.error(f"Error msg: {err_msg}")
        self.run_pipeline_end()
        self.set_viewer_state_to_stop()
        if "FileNotFoundError" in exc_msg:
            showerror(
                "Pipeline Runtime Error",
                f"Missing pipeline file:\n\n{self.pipeline_path}",
            )
        if "CUDA" in exc_msg:
            showerror("CUDA Runtime Error", "Please see logs for more details")

    def run_pipeline_end(self) -> None:
        """Called when pipeline execution is completed.
        To perform clean-up/housekeeping tasks to ensure system consistency"""
        self.logger.debug("run pipeline end")
        for node in self._pipeline.nodes:
            if node.name.endswith("input.visual"):
                node.release_resources()  # clean up nodes with threads
        self.is_pipeline_running = False
        self._enable_slider()
        self.set_viewer_state_to_stop()
        self._set_header_stop()

    def run_pipeline_one_iteration(self) -> None:  # pylint: disable=too-many-branches
        """Run one pipeline iteration"""
        self.is_pipeline_running = True
        err_runtime = False
        exc_msg = ""
        err_stream = StringIO()
        # technote: Detect runtime exception with flag as exception object holds ref to
        # error stack frame, preventing further objects from being freed.
        with redirect_stderr(err_stream):
            try:
                for node in self._pipeline.nodes:
                    if self._pipeline.data.get("pipeline_end", False):
                        self._pipeline.terminate = True
                        if "pipeline_end" not in node.inputs:
                            continue
                    if "all" in node.inputs:
                        inputs = copy.deepcopy(self._pipeline.data)
                    else:
                        inputs = {
                            key: self._pipeline.data[key]
                            for key in node.inputs
                            if key in self._pipeline.data
                        }
                    if hasattr(node, "optional_inputs"):
                        # Nodes won't receive inputs with optional key if not found upstream
                        for key in node.optional_inputs:
                            if key in self._pipeline.data:
                                inputs[key] = self._pipeline.data[key]
                    if node.name.endswith("output.screen"):
                        pass  # disable duplicate video from output.screen
                    else:
                        outputs = node.run(inputs)
                        self._pipeline.data.update(outputs)
                    # check for FPS on first iteration
                    if self._frame_idx == 0 and node.name.endswith("input.visual"):
                        num_frames = node.total_frame_count
                        if num_frames > 0:
                            self.tk_progress["maximum"] = num_frames
                        else:
                            self.tk_progress["mode"] = "indeterminate"
            except Exception:  # pylint: disable=broad-except
                err_runtime = True
                exc_msg = traceback.format_exc()

        # handle pipeline runtime error
        if err_runtime:
            self.pipeline_error(exc_msg, err_stream)
            return

        # render img into screen output to Tkinter
        img = self._pipeline.data["img"]
        if img is not None:
            frame = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)  # BGR -> RGB for Tkinter
            self._frames.append(frame)  # save frame for playback
            self._frame_idx += 1
            self._show_frame()

        # update progress bar after each iteration
        self.tk_progress["value"] = self._frame_idx
        self.tk_lbl_frame_num["text"] = self._frame_idx + 1

        # check if need to stop after fixed number of iterations
        if self.num_iter and self._frame_idx + 1 >= self.num_iter:
            self.logger.info(f"Stopping pipeline after {self.num_iter} iterations")
            self.stop_running_pipeline()

    def run_pipeline_start(self) -> None:
        """Start PeekingDuck's pipeline"""
        self.logger.debug("run pipeline start")
        self.logger.debug(f"pipeline path: {self.pipeline_path}")
        self.logger.debug(f"custom_nodes: {self.custom_nodes_parent_path}")
        # technotes: node __init__() is called in get_pipeline() below.
        #            Detect runtime exception with flag and handle it outside as
        #            exception object holds reference to stack frame where error was
        #            raised, preventing memory from being freed.
        err_runtime = False
        exc_msg = ""
        err_stream = StringIO()
        with redirect_stderr(err_stream):
            try:
                self._node_loader = DeclarativeLoader(
                    self.pipeline_path,
                    self.config_updates_cli,
                    self.custom_nodes_parent_path,
                    pkd_viewer=True,
                )
                self._pipeline: Pipeline = self._node_loader.get_pipeline()
            except Exception:  # pylint: disable=broad-except
                err_runtime = True
                exc_msg = traceback.format_exc()

        # handle pipeline runtime error
        if err_runtime:
            self.pipeline_error(exc_msg, err_stream)
        else:
            self._set_header_running()
            self.set_viewer_state_to_play()
            self._set_status_text(f"Pipeline: {self.pipeline_full_path}")
            self.is_pipeline_running = True
            self._enable_progress()

    def stop_running_pipeline(self) -> None:
        """Signal pipeline execution to be stopped"""
        self._pipeline.terminate = True

    ###################
    #
    # Pipeline Playback
    #
    ###################
    def start_playback(self) -> None:
        """Start output playback process"""
        self.is_output_playback = True
        # auto-rewind if at last frame
        if self._frame_idx + 1 >= len(self._frames):
            self._frame_idx = 0
            self._update_slider_and_show_frame()
        self.set_viewer_state_to_play()
        self._set_header_playing()
        self.do_playback()

    def do_playback(self) -> None:
        """Playback saved video frames: to be called continuously"""
        if self._forward_one_frame():
            self.tk_scale.set(self._frame_idx + 1)
        else:
            self.stop_playback()

    def stop_playback(self) -> None:
        """Stop output playback"""
        self.is_output_playback = False
        self.set_viewer_state_to_stop()
        self._set_header_stop()

    def set_viewer_state_to_play(self) -> None:
        """Set self state to play for either 1) pipeline execution or 2) playback"""
        self.state = "play"
        self.tk_btn_play["text"] = STOP_BUTTON_TEXT

    def set_viewer_state_to_stop(self) -> None:
        """Set self state to stop"""
        self.state = "stop"
        self.tk_btn_play["text"] = PLAY_BUTTON_TEXT



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Detection Utilities for Region Proposal Network and Region of Interest Operations
Modifications include:
- Removed unused methods from BoxCoder class
    - encode
    - encode_single
    - encode_boxes
"""

from typing import List, Tuple
import math
import torch
from torch import Tensor


class BoxCoder:
    """
    This class decodes a set of bounding boxes into
    the representation used for training the regressors.
    """

    def __init__(
        self,
        weights: Tuple[float, float, float, float],
        bbox_xform_clip: float = math.log(1000.0 / 16),
    ) -> None:
        """
        Args:
            weights (4-element tuple): Each element corresponds to the weights that the raw
                regressed offsets (dx, dy, dw, dh) will be divided by. (i.e. the higher the weight,
                the less effect the regressed offsets have on the boxes)
            bbox_xform_clip (float): The log of maximum allowable width or height scale offset.
                E.g. if maximum allowable width and height offset is 3 times of original width or
                height, the bbox_xform_clip will be log(3). It is also to prevent sending too large
                an input to torch.exp() when calculating the predicted width and height.
        """
        self.weights = weights
        self.bbox_xform_clip = bbox_xform_clip

    def decode(self, rel_codes: Tensor, boxes: List[Tensor]) -> Tensor:
        """
        From a list of original boxes and encoded relative box offsets,
        get the decoded boxes.

        Args:
            rel_codes (Tensor): encoded boxes offsets regressions
            boxes (List[Tensor]): List of reference boxes.

        Returns:
            Tensor: Decoded boxes
        """
        assert isinstance(boxes, (list, tuple))
        assert isinstance(rel_codes, Tensor)
        boxes_per_image = [b.size(0) for b in boxes]
        concat_boxes = torch.cat(boxes, dim=0)
        box_sum = 0
        for val in boxes_per_image:
            box_sum += val
        if box_sum > 0:
            rel_codes = rel_codes.reshape(box_sum, -1)
        pred_boxes = self.decode_single(rel_codes, concat_boxes)
        if box_sum > 0:
            pred_boxes = pred_boxes.reshape(box_sum, -1, 4)
        return pred_boxes

    def decode_single(self, rel_codes: Tensor, boxes: Tensor) -> Tensor:
        # pylint: disable=too-many-locals
        """Performs decoding operation for the boxes

        Args:
            rel_codes (Tensor): encoded boxes offsets regressions
            boxes (Tensor): Reference boxes

        Returns:
            Tensor: Decoded boxes
        """
        boxes = boxes.to(rel_codes.dtype)

        widths = boxes[:, 2] - boxes[:, 0]
        heights = boxes[:, 3] - boxes[:, 1]
        ctr_x = boxes[:, 0] + 0.5 * widths
        ctr_y = boxes[:, 1] + 0.5 * heights

        weight_x, weight_y, weight_w, weight_h = self.weights
        del_x = rel_codes[:, 0::4] / weight_x
        del_y = rel_codes[:, 1::4] / weight_y
        del_w = rel_codes[:, 2::4] / weight_w
        del_h = rel_codes[:, 3::4] / weight_h

        # Prevent sending too large values into torch.exp()
        del_w = torch.clamp(del_w, max=self.bbox_xform_clip)
        del_h = torch.clamp(del_h, max=self.bbox_xform_clip)

        pred_ctr_x = del_x * widths[:, None] + ctr_x[:, None]
        pred_ctr_y = del_y * heights[:, None] + ctr_y[:, None]
        pred_w = torch.exp(del_w) * widths[:, None]
        pred_h = torch.exp(del_h) * heights[:, None]

        # Distance from center to box's corner.
        c_to_c_h = (
            torch.tensor(0.5, dtype=pred_ctr_y.dtype, device=pred_h.device) * pred_h
        )
        c_to_c_w = (
            torch.tensor(0.5, dtype=pred_ctr_x.dtype, device=pred_w.device) * pred_w
        )

        pred_boxes1 = pred_ctr_x - c_to_c_w
        pred_boxes2 = pred_ctr_y - c_to_c_h
        pred_boxes3 = pred_ctr_x + c_to_c_w
        pred_boxes4 = pred_ctr_y + c_to_c_h
        pred_boxes = torch.stack(
            (pred_boxes1, pred_boxes2, pred_boxes3, pred_boxes4), dim=2
        ).flatten(1)
        return pred_boxes



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Creates a zone from a polygon area
"""

from typing import List, Tuple

from shapely.geometry.polygon import Point, Polygon


class Zone:  # pylint: disable=too-few-public-methods
    """This class uses polygon area to create a zone for counting."""

    def __init__(self, points: List[List[int]]) -> None:
        # Each zone is a polygon created by a list of x, y coordinates
        self.polygon_points = [tuple(point) for point in points]
        self.polygon = Polygon(self.polygon_points)

    def contains(self, point: Tuple[float, float]) -> bool:
        """Checks if the specified point is within the area bounded by the
        zone.

        Args:
            point (Tuple[float, float]): The x- and y- coordinate to check.

        Returns:
            (bool): True if ``point`` is within the zone.
        """
        point = Point(point)
        return self.polygon.buffer(1).contains(point)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Implements the Mask R-CNN model.
Modifications include:
- Removed training related arguments for ROI_Heads and __init__:
    - box_fg_iou_thresh
    - box_bg_iou_thresh
    - box_batch_size_per_image
    - box_positive_fraction
- Removed training related arguments for RegionProposalNetwork and __init__:
    - fg_iou_thresh
    - bg_iou_thresh
    - batch_size_per_image
    - positive_fraction
    - rpn_pre_nms_top_n_train
    - rpn_post_nms_top_n_train
"""

from typing import Iterable, Optional, Tuple
from collections import OrderedDict
from torch import nn
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.ops import poolers
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detection import (
    faster_rcnn,
)


class MaskRCNN(faster_rcnn.FasterRCNN):
    """
    Implements Mask R-CNN.

    The input to the model is expected to be a list of tensors, each of shape [C, H, W], one for
    each image, and should be in 0-1 range. Different images can have different sizes.

    During inference, the model requires only the input tensors, and returns the post-processed
    predictions as a List[Dict[Tensor]], one for each input image. The fields of the Dict are as
    follows:
        - boxes (``FloatTensor[N, 4]``): the predicted boxes in ``[x1, y1, x2, y2]`` format, with
          ``0 <= x1 < x2 <= W`` and ``0 <= y1 < y2 <= H``.
        - labels (Int64Tensor[N]): the predicted labels for each image
        - scores (Tensor[N]): the scores of each prediction
        - masks (UInt8Tensor[N, 1, H, W]): the predicted masks for each instance, in 0-1 range.
          In order to obtain the final segmentation masks, the soft masks can be thresholded,
          generally with a value of 0.5 (mask >= 0.5)

    Args:
        backbone (nn.Module): the network used to compute the features for the model.
            It should contain a out_channels attribute, which indicates the number of output
            channels that each feature map has (and it should be the same for all feature maps).
            The backbone should return a single Tensor or and OrderedDict[Tensor].
        num_classes (int): number of output classes of the model (including the background).
            If box_predictor is specified, num_classes should be None.
        min_size (int): minimum size of the image to be rescaled before feeding it to the backbone
        max_size (int): maximum size of the image to be rescaled before feeding it to the backbone
        image_mean (Optional[Iterable[float]]): mean values used for input normalization.
            They are generally the mean values of the dataset on which the backbone has been
            trained on. Requires 3 floating point elements in the Iterable
        image_std (Optional[Iterable[float]]): std values used for input normalization.
            They are generally the std values of the dataset on which the backbone has been trained
            on. Requires 3 floating point elements in the Iterable
        rpn_anchor_generator (AnchorGenerator): module that generates the anchors for a set of
            feature maps.
        rpn_head (nn.Module): module that computes the objectness and regression deltas from the
            RPN
        rpn_pre_nms_top_n_test (int): number of proposals to keep before applying NMS during
            testing
        rpn_post_nms_top_n_test (int): number of proposals to keep after applying NMS during
            testing
        rpn_nms_thresh (float): NMS threshold used for postprocessing the RPN proposals
        rpn_score_thresh (float): during inference, only return proposals with a classification
            score greater than rpn_score_thresh
        box_roi_pool (Optional[nn.Module]): the module which crops and resizes the feature maps in
            the locations indicated by the bounding boxes
        box_head (Optional[nn.Module]): module that takes the cropped feature maps as input
        box_predictor (Optional[nn.Module]): module that takes the output of box_head and returns
            the classification logits and box regression deltas.
        box_score_thresh (float): during inference, only return proposals with a classification
            score greater than box_score_thresh
        box_nms_thresh (float): NMS threshold for the prediction head. Used during inference
        box_detections_per_img (int): maximum number of detections per image, for all classes.
        bbox_reg_weights (Optional[Tuple[float, float, float, float]]): weights for the
            encoding/decoding of the bounding boxes
        mask_roi_pool (Optional[nn.Module]): the module which crops and resizes the feature maps in
            the locations indicated by the bounding boxes, which will be used for the mask head.
        mask_head (Optional[nn.Module]): module that takes the cropped feature maps as input
        mask_predictor (Optional[nn.Module]): module that takes the output of the mask_head and
            returns the segmentation mask logits
    """

    # pylint: disable=too-many-arguments,too-many-locals
    def __init__(
        self,
        backbone: nn.Module,
        num_classes: Optional[int] = None,
        # transform parameters
        min_size: int = 800,
        max_size: int = 1333,
        image_mean: Optional[Iterable[float]] = None,
        image_std: Optional[Iterable[float]] = None,
        # RPN parameters
        rpn_anchor_generator: Optional[nn.Module] = None,
        rpn_head: Optional[nn.Module] = None,
        rpn_pre_nms_top_n_test: int = 1000,
        rpn_post_nms_top_n_test: int = 1000,
        rpn_nms_thresh: float = 0.7,
        rpn_score_thresh: float = 0.0,
        # Box parameters
        box_roi_pool: Optional[nn.Module] = None,
        box_head: Optional[nn.Module] = None,
        box_predictor: Optional[nn.Module] = None,
        box_score_thresh: float = 0.05,
        box_nms_thresh: float = 0.5,
        box_detections_per_img: int = 100,
        bbox_reg_weights: Optional[Tuple[float, float, float, float]] = None,
        # Mask parameters
        mask_roi_pool: Optional[nn.Module] = None,
        mask_head: Optional[nn.Module] = None,
        mask_predictor: Optional[nn.Module] = None,
    ):

        assert isinstance(mask_roi_pool, (poolers.MultiScaleRoIAlign, type(None)))

        if num_classes is not None:
            if mask_predictor is not None:
                raise ValueError(
                    "num_classes should be None when mask_predictor is specified"
                )

        out_channels = backbone.out_channels

        if mask_roi_pool is None:
            mask_roi_pool = poolers.MultiScaleRoIAlign(
                featmap_names=["0", "1", "2", "3"], output_size=14, sampling_ratio=2
            )

        if mask_head is None:
            mask_layers = (256, 256, 256, 256)
            mask_dilation = 1
            mask_head = MaskRCNNHeads(
                out_channels, mask_layers, mask_dilation  # type: ignore[arg-type]
            )

        if mask_predictor is None:
            mask_predictor_in_channels = 256  # == mask_layers[-1]
            mask_dim_reduced = 256
            mask_predictor = MaskRCNNPredictor(
                mask_predictor_in_channels, mask_dim_reduced, num_classes  # type: ignore[arg-type]
            )

        super().__init__(
            backbone,
            num_classes,
            # transform parameters
            min_size,
            max_size,
            image_mean,
            image_std,
            # RPN-specific parameters
            rpn_anchor_generator,
            rpn_head,
            rpn_pre_nms_top_n_test,
            rpn_post_nms_top_n_test,
            rpn_nms_thresh,
            rpn_score_thresh,
            # Box parameters
            box_roi_pool,
            box_head,
            box_predictor,
            box_score_thresh,
            box_nms_thresh,
            box_detections_per_img,
            bbox_reg_weights,
        )

        self.roi_heads.mask_roi_pool = mask_roi_pool
        self.roi_heads.mask_head = mask_head
        self.roi_heads.mask_predictor = mask_predictor


class MaskRCNNHeads(nn.Sequential):
    """Implements the head for Mask R-CNN, a module that takes the cropped feature maps as input"""

    def __init__(self, in_channels: int, layers: Iterable[int], dilation: int):
        """
        Args:
            in_channels (int): number of input channels
            layers (Iterable[int]): feature dimensions of each FCN layer
            dilation (int): dilation rate of kernel
        """
        layers_dict: "OrderedDict[str, nn.Module]" = OrderedDict()
        next_feature = in_channels
        for layer_idx, layer_features in enumerate(layers, 1):
            layers_dict["mask_fcn{}".format(layer_idx)] = nn.Conv2d(
                next_feature,
                layer_features,
                kernel_size=3,
                stride=1,
                padding=dilation,
                dilation=dilation,
            )
            layers_dict["relu{}".format(layer_idx)] = nn.ReLU(inplace=True)
            next_feature = layer_features

        super().__init__(layers_dict)
        for name, param in self.named_parameters():
            if "weight" in name:
                nn.init.kaiming_normal_(param, mode="fan_out", nonlinearity="relu")


class MaskRCNNPredictor(nn.Sequential):
    """A module that takes the output of the mask_head and returns the
    segmentation mask logits"""

    def __init__(self, in_channels: int, dim_reduced: int, num_classes: int):
        super().__init__(
            OrderedDict(
                [
                    (
                        "conv5_mask",
                        nn.ConvTranspose2d(in_channels, dim_reduced, 2, 2, 0),
                    ),
                    ("relu", nn.ReLU(inplace=True)),
                    ("mask_fcn_logits", nn.Conv2d(dim_reduced, num_classes, 1, 1, 0)),
                ]
            )
        )

        for name, param in self.named_parameters():
            if "weight" in name:
                nn.init.kaiming_normal_(param, mode="fan_out", nonlinearity="relu")



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# MIT License

# Copyright (c) 2020 Haotian Liu and Rafael A. Rivera-Soto

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Backbone for YolactEdge.
Modifications include:
- Removed unused ResNetBackboneGN class
- Removed unused darknetconvlayer
- Removed unused DarkNetBlock class
- Removed unused DarkNetBackbone class
- Removed unused VGGBackBone class
- Refactor and formatting
"""

from typing import Callable, List, Any, Tuple, Union, Optional
from functools import partial
import torch.nn as nn
from torch import Tensor


class Bottleneck(nn.Module):  # pylint: disable=too-many-instance-attributes
    """Bottleneck in torchvision places the stride for downsampling at 3x3
    convolution(self.conv2) while original implementation places the stride at
    the first 1x1 convolution(self.conv1) according to "Deep residual learning
    for image recognition"https://arxiv.org/abs/1512.03385.

    Adapted from torchvision.models.resnet
    """

    expansion = 4

    def __init__(  # pylint: disable=too-many-arguments
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Callable = nn.Sequential(),
        norm_layer: Callable = nn.BatchNorm2d,
        dilation: int = 1,
    ) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(
            inplanes, planes, kernel_size=1, bias=False, dilation=dilation
        )
        self.bn1 = norm_layer(planes)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            bias=False,
            dilation=dilation,
        )
        self.bn2 = norm_layer(planes)
        self.conv3 = nn.Conv2d(
            planes, planes * 4, kernel_size=1, bias=False, dilation=dilation
        )
        self.bn3 = norm_layer(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = nn.Sequential() if downsample is None else downsample
        self.stride = stride

    def forward(self, inputs: Tensor) -> Tensor:
        """Forward propagation of the Bottleneck blocks of ResNet.

        Args:
            inputs (Tensor): Input tensor

        Returns:
            out (Tensor): Output tensor
        """
        out = self.conv1(inputs)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        residual = self.downsample(inputs)
        out += residual
        out = self.relu(out)

        return out


class ResNetBackbone(nn.Module):  # pylint: disable=too-many-instance-attributes
    """Implements ResNet backbone, adapted from torchvision.models.resnet"""

    def __init__(
        self,
        layers: List[int],
        block: Callable = Bottleneck,
        norm_layer: Callable = nn.BatchNorm2d,
    ) -> None:
        super().__init__()
        self.num_base_layers = len(layers)
        self.layers = nn.ModuleList()
        self.channels: List[Any] = []
        self.norm_layer = norm_layer
        self.dilation = 1
        self.atrous_layers: List[Any] = []

        self.inplanes = 64

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = norm_layer(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self._make_layer(block, 64, layers[0])
        self._make_layer(block, 128, layers[1], stride=2)
        self._make_layer(block, 256, layers[2], stride=2)
        self._make_layer(block, 512, layers[3], stride=2)

        self.backbone_modules = [m for m in self.modules() if isinstance(m, nn.Conv2d)]

    def _make_layer(
        self, block: Callable, planes: int, blocks: int, stride: int = 1
    ) -> Callable:
        """Method to make layers for ResNet"""
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            if len(self.layers) in self.atrous_layers:
                self.dilation += 1
                stride = 1

            downsample = nn.Sequential(
                nn.Conv2d(
                    self.inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                    dilation=self.dilation,
                ),
                self.norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.norm_layer,
                self.dilation,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, norm_layer=self.norm_layer))

        layer = nn.Sequential(*layers)
        self.channels.append(planes * block.expansion)
        self.layers.append(layer)
        return layer

    def forward(self, inputs: Tensor, partial_bn: bool = False) -> List[Tensor]:
        """Forward propoagation of the main ResNet model that returns a list of
        convouts for each layer.

        Args:
            inputs (Tensor): Input Tensor

        Returns:
            outs (Tensor): Output Tensor
        """
        inputs = self.conv1(inputs)
        inputs = self.bn1(inputs)
        inputs = self.relu(inputs)
        inputs = self.maxpool(inputs)

        outs = []
        layer_idx = 0
        for layer in self.layers:
            layer_idx += 1
            if not partial_bn or layer_idx <= 2:
                inputs = layer(inputs)
                outs.append(inputs)
        return outs

    def add_layer(
        self,
        conv_channels: int = 1024,
        downsample: int = 2,
        depth: int = 1,
        block: Callable = Bottleneck,
    ) -> None:
        """Add a downsample layer to the backbone as per what SSD does.

        Args:
            conv_channels (int): Number of channels in the convolution.
            downsample (int): Number of downsampling layers.
            depth (int): Number of blocks in the downsample layer.
            block: Block to use for the downsample layer.
        """
        self._make_layer(
            block, conv_channels // block.expansion, blocks=depth, stride=downsample
        )


class ConvBNAct(nn.Sequential):
    """Adapted from torchvision.models.mobilenet.ConvBNReLU"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        in_planes: int,
        out_planes: int,
        kernel_size: int = 3,
        stride: int = 1,
        groups: int = 1,
        activation: Any = nn.ReLU6(inplace=True),
    ) -> None:
        padding = (kernel_size - 1) // 2
        super().__init__(
            nn.Conv2d(
                in_planes,
                out_planes,
                kernel_size,
                stride,
                padding,
                groups=groups,
                bias=False,
            ),
            nn.BatchNorm2d(out_planes),
            activation,
        )


ConvBNReLU = partial(ConvBNAct)


class InvertedResidual(nn.Module):
    """Residual block that uses an inverted structure for efficiency for
    mobile-optimized CNNs as specified in the <https://arxiv.org/abs/1801.04381>
    paper, adapted from torchvision.models.mobilenet.InvertedResidual"""

    def __init__(self, inp: int, oup: int, stride: int, expand_ratio: int) -> None:
        super().__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = int(inp * expand_ratio)
        self.use_res_connect = self.stride == 1 and inp == oup

        layers: List[Union[ConvBNAct, nn.Conv2d, nn.BatchNorm2d]] = []
        if expand_ratio != 1:
            layers.append(ConvBNReLU(inp, hidden_dim, kernel_size=1))

        layers.extend(
            [
                ConvBNReLU(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim),
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            ]
        )
        self.conv = nn.Sequential(*layers)

    def forward(self, inputs: Tensor) -> Tensor:
        """Returns a convout for the inverted residual block.

        Args:
            inputs (Tensor): Input Tensor

        Returns:
            Tensor: Convout for the inverted residual block
        """
        if self.use_res_connect:
            return inputs + self.conv(inputs)
        return self.conv(inputs)


class MobileNetV2Backbone(nn.Module):
    """MobileNet V2 main class adapted from torchvision.models.mobilenet.MobileNetV2
    from the `MobileNetV2: Inverted Residuals and Linear Bottlenecks
    <https://arxiv.org/abs/1801.04381>` paper.

    Args:
        width_mult (float): Width multiplier - adjusts number of channels in each
            layer by this amount inverted_residual_setting: Network structure
        inverted_residual_setting: Network structure
        round_nearest (int): Round the number of channels in each layer to be a
            multiple of this number. Set to 1 to turn off rounding.
        block: Module specifying inverted residual builing block for mobilenet
    """

    def __init__(
        self,
        width_mult: float = 1.0,
        inverted_residual_setting: Any = None,
        round_nearest: int = 8,
        block: Callable = InvertedResidual,
    ) -> None:
        super().__init__()

        input_channel = 32
        last_channel = 1280
        self.channels = []
        self.layers = nn.ModuleList()

        if inverted_residual_setting is None:
            raise ValueError(
                "Must provide inverted_residual_setting where each "
                "element is a list that represents the "
                "MobileNetV2 t,c,n,s values for that layer."
            )
        if (
            len(inverted_residual_setting) == 0
            or len(inverted_residual_setting[0]) != 4
        ):
            raise ValueError(
                "inverted_residual_setting should be non-empty "
                "or a 4-element list, got {}".format(inverted_residual_setting)
            )

        input_channel = self._make_divisible(input_channel * width_mult, round_nearest)
        self.last_channel = self._make_divisible(
            last_channel * max(1.0, width_mult), round_nearest
        )
        self.layers.append(ConvBNReLU(3, input_channel, stride=2))
        self.channels.append(input_channel)

        for t_val, c_val, n_val, s_val in inverted_residual_setting:
            input_channel = self._make_layer(
                input_channel,
                width_mult,
                round_nearest,
                t_val,
                c_val,
                n_val,
                s_val,
                block,
            )
            self.channels.append(input_channel)

        self.layers.append(ConvBNReLU(input_channel, self.last_channel, kernel_size=1))
        self.channels.append(self.last_channel)
        self.backbone_modules = [m for m in self.modules() if isinstance(m, nn.Conv2d)]

    def _make_layer(  # pylint: disable=too-many-arguments
        self,
        input_channel: int,
        width_mult: float,
        round_nearest: int,
        t_val: int,
        c_val: int,
        n_val: int,
        s_val: int,
        block: Callable,
    ) -> int:
        """A layer is a combination of stacked InvertedResidual blocks to build
        a layer for MobileNetV2.

        Args:
            input_channel (int): The number of input channels to the first block.
            width_mult (float): The width multiplier for the model. Adjusts number
                of channels in each layer by this amount.
            round_nearest (int): Round the number of channels in each layer to be
                a multiple of this number. Set to 1 to turn off rounding
            t_val (int): expansion factor
            c_val (int): output channels
            n_val (int): number of repetitions
            s_val (int): stride of the first layer of each sequence
            block (Callable): Module specifying inverted residual building block
                for MobileNetV2

        Returns:
            input_channel (int): The number of input channels to the next block.
        """
        layers = []
        output_channel = self._make_divisible(c_val * width_mult, round_nearest)

        for i in range(n_val):
            stride = s_val if i == 0 else 1
            layers.append(
                block(input_channel, output_channel, stride, expand_ratio=t_val)
            )
            input_channel = output_channel

        self.layers.append(nn.Sequential(*layers))
        return input_channel

    def forward(self, inputs: Tensor) -> Tuple[Tensor, ...]:
        """Forward propoagation of the MobileNet V2 model that returns a Tuple of
        convouts for each layer.

        Args:
            inputs (Tensor): Input Tensor

        Returns:
            outs (Tuple[Tensor, ...]): Output Tensor
        """
        outs = []

        for _, layer in enumerate(self.layers):
            inputs = layer(inputs)
            outs.append(inputs)

        return tuple(outs)

    def add_layer(  # pylint: disable=too-many-arguments
        self,
        conv_channels: int = 1280,
        t_val: int = 1,
        c_val: int = 1280,
        n_val: int = 1,
        s_val: int = 2,
    ) -> None:
        """
        Args:
            conv_channels (int): number of channels in the convout of the previous layer
            t_val (int): expansion factor
            c_val (int): output channels
            n_val (int): number of repetitions
            s_val (int): stride of the first layer of each sequence. All others
                use a stride of 1
        """
        self._make_layer(
            conv_channels, 1.0, 8, t_val, c_val, n_val, s_val, InvertedResidual
        )

    @classmethod
    def _make_divisible(
        cls, value: float, divisor: int, min_value: Optional[int] = None
    ) -> int:
        """Adapted from torchvision.models.mobilenet._make_divisable
        This function  is taken from the original tf repo. It ensures that all
        layers have a channel number that is divisible by 8

        Args:
            value (float): The input number to make it divisible by 8
            divisor (int): The divisor to use
            min_value (Any): The minimum value to use

        Returns:
            new_value (int): The new value that is divisible by 8
        """
        if min_value is None:
            min_value = divisor
        new_value = max(min_value, int(value + divisor / 2) // divisor * divisor)
        # Make sure that round down does not go down by more than 10%.
        if new_value < 0.9 * value:
            new_value += divisor
        return new_value



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original Copyright From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America
#           (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute
#           (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Implements Faster R-CNN model from https://arxiv.org/pdf/1506.01497.pdf.
Modifications include:
- Removed training related arguments for ROI_Heads and __init__:
    - box_fg_iou_thresh
    - box_bg_iou_thresh
    - box_batch_size_per_image
    - box_positive_fraction
- Removed training related arguments for RegionProposalNetwork and __init__:
    - fg_iou_thresh
    - bg_iou_thresh
    - batch_size_per_image
    - positive_fraction
    - rpn_pre_nms_top_n_train
    - rpn_post_nms_top_n_train
"""

from typing import Optional, Tuple, Iterable
from torch import nn, Tensor
import torch.nn.functional as F
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.ops import poolers
from peekingduck.pipeline.nodes.model.mask_rcnnv1.mask_rcnn_files.detection import (
    anchor_utils,
    rpn,
    roi_heads,
    generalized_rcnn as g_rnn,
    transform as g_rcnn_transform,
)


class FasterRCNN(g_rnn.GeneralizedRCNN):
    """
    Implements Faster R-CNN.

    The input to the model is expected to be a list of tensors, each of shape [C, H, W], one
    for each image, and should be in 0-1 range. Different images can have different sizes.

    During inference, the model requires only the input tensors, and returns the
    post-processed predictions as a List[Dict[Tensor]], one for each input image. The fields
    of the Dict are as follows:
        - boxes (``FloatTensor[N, 4]``): the predicted boxes in ``[x1, y1, x2, y2]`` format,
          with ``0 <= x1 < x2 <= W`` and ``0 <= y1 < y2 <= H``.
        - labels (Int64Tensor[N]): the predicted labels for each image
        - scores (Tensor[N]): the scores or each prediction

    Args:
        backbone (nn.Module): the network used to compute the features for the model.
            It should contain a out_channels attribute, which indicates the number of output
            channels that each feature map has (and it should be the same for all feature maps).
            The backbone should return a single Tensor or and OrderedDict[Tensor].
        num_classes (int): number of output classes of the model (including the background).
            If box_predictor is specified, num_classes should be None.
        min_size (int): minimum size of the image to be rescaled before feeding it to the backbone
        max_size (int): maximum size of the image to be rescaled before feeding it to the backbone
        image_mean (Optional[Iterable[float]]): mean values used for input normalization.
            They are generally the mean values of the dataset on which the backbone has been
            trained on
        image_std (Optional[Iterable[float]]): std values used for input normalization.
            They are generally the std values of the dataset on which the backbone has been trained
            on
        rpn_anchor_generator (Optional[nn.Module]): module that generates the anchors for a set of
            feature maps.
        rpn_head (Optional[nn.Module]): module that computes the objectness and regression deltas
            from the RPN
        rpn_pre_nms_top_n_test (int): number of proposals to keep before applying NMS during
            testing
        rpn_post_nms_top_n_test (int): number of proposals to keep after applying NMS during
            testing
        rpn_nms_thresh (float): NMS threshold used for postprocessing the RPN proposals
        rpn_score_thresh (float): during inference, only return proposals with a classification
            score greater than rpn_score_thresh
        box_roi_pool (Optional[nn.Module]): the module which crops and resizes the feature maps in
            the locations indicated by the bounding boxes
        box_head (Optional[nn.Module]): module that takes the cropped feature maps as input
        box_predictor (Optional[nn.Module]): module that takes the output of box_head and returns
            the classification logits and box regression deltas.
        box_score_thresh (float): during inference, only return proposals with a classification
            score greater than box_score_thresh
        box_nms_thresh (float): NMS threshold for the prediction head. Used during inference
        box_detections_per_img (int): maximum number of detections per image, for all classes.
        bbox_reg_weights (Optional[Tuple[float, float, float, float]]): weights for the
            encoding/decoding of the bounding boxes
    """

    # pylint: disable=too-many-arguments,too-many-locals
    def __init__(
        self,
        backbone: nn.Module,
        num_classes: Optional[int] = None,
        # transform parameters
        min_size: int = 800,
        max_size: int = 1333,
        image_mean: Optional[Iterable[float]] = None,
        image_std: Optional[Iterable[float]] = None,
        # RPN parameters
        rpn_anchor_generator: Optional[nn.Module] = None,
        rpn_head: Optional[nn.Module] = None,
        rpn_pre_nms_top_n_test: int = 1000,
        rpn_post_nms_top_n_test: int = 1000,
        rpn_nms_thresh: float = 0.7,
        rpn_score_thresh: float = 0.0,
        # Box parameters
        box_roi_pool: Optional[nn.Module] = None,
        box_head: Optional[nn.Module] = None,
        box_predictor: Optional[nn.Module] = None,
        box_score_thresh: float = 0.05,
        box_nms_thresh: float = 0.5,
        box_detections_per_img: int = 100,
        bbox_reg_weights: Optional[Tuple[float, float, float, float]] = None,
    ):

        if not hasattr(backbone, "out_channels"):
            raise ValueError(
                "backbone should contain an attribute out_channels "
                "specifying the number of output channels (assumed to be the "
                "same for all the levels)"
            )

        assert isinstance(
            rpn_anchor_generator, (anchor_utils.AnchorGenerator, type(None))
        )
        assert isinstance(box_roi_pool, (poolers.MultiScaleRoIAlign, type(None)))

        if num_classes is not None:
            if box_predictor is not None:
                raise ValueError(
                    "num_classes should be None when box_predictor is specified"
                )
        else:
            if box_predictor is None:
                raise ValueError(
                    "num_classes should not be None when box_predictor "
                    "is not specified"
                )

        out_channels = backbone.out_channels

        if rpn_anchor_generator is None:
            anchor_sizes = ((32,), (64,), (128,), (256,), (512,))
            aspect_ratios = ((0.5, 1.0, 2.0),) * len(anchor_sizes)
            rpn_anchor_generator = anchor_utils.AnchorGenerator(
                anchor_sizes, aspect_ratios
            )
        if rpn_head is None:
            rpn_head = rpn.RPNHead(
                out_channels,  # type: ignore[arg-type]
                rpn_anchor_generator.num_anchors_per_location()[0],
            )

        rpn_pre_nms_top_n = dict(testing=rpn_pre_nms_top_n_test)
        rpn_post_nms_top_n = dict(testing=rpn_post_nms_top_n_test)

        rpn_obj = rpn.RegionProposalNetwork(
            rpn_anchor_generator,
            rpn_head,
            rpn_pre_nms_top_n,
            rpn_post_nms_top_n,
            rpn_nms_thresh,
            score_thresh=rpn_score_thresh,
        )

        if box_roi_pool is None:
            box_roi_pool = poolers.MultiScaleRoIAlign(
                featmap_names=["0", "1", "2", "3"], output_size=7, sampling_ratio=2
            )

        if box_head is None:
            resolution = box_roi_pool.output_size[0]
            representation_size = 1024
            box_head = TwoMLPHead(
                out_channels * resolution ** 2,  # type: ignore[operator,arg-type]
                representation_size,
            )

        if box_predictor is None:
            representation_size = 1024
            box_predictor = FastRCNNPredictor(
                representation_size, num_classes  # type: ignore[arg-type]
            )

        roi_heads_obj = roi_heads.RoIHeads(
            # Box
            box_roi_pool,
            box_head,
            box_predictor,
            bbox_reg_weights,
            box_score_thresh,
            box_nms_thresh,
            box_detections_per_img,
        )

        if image_mean is None:
            image_mean = [0.485, 0.456, 0.406]
        if image_std is None:
            image_std = [0.229, 0.224, 0.225]
        transform = g_rcnn_transform.GeneralizedRCNNTransform(
            min_size, max_size, image_mean, image_std
        )

        super().__init__(backbone, rpn_obj, roi_heads_obj, transform)


class TwoMLPHead(nn.Module):
    """Standard heads for FPN-based models. Mainly used to perform transformation on the cropped
    features from the RoI pooling function

    Args:
        in_channels (int): number of input channels
        representation_size (int): size of the intermediate representation
    """

    def __init__(self, in_channels: int, representation_size: int):
        super().__init__()

        self.fc6 = nn.Linear(in_channels, representation_size)
        self.fc7 = nn.Linear(representation_size, representation_size)

    def forward(self, inputs: Tensor) -> Tensor:
        """Takes in input and perform forward propagation through 2 layers of fully-connected
        layers.

        Args:
            inputs (Tensor): Input Tensor

        Returns:
            Tensor: Output Tensor
        """
        inputs = inputs.flatten(start_dim=1)

        inputs = F.relu(self.fc6(inputs))
        x_out = F.relu(self.fc7(inputs))

        return x_out


class FastRCNNPredictor(nn.Module):
    """
    Standard classification + bounding box regression layers
    for Fast R-CNN.

    Args:
        in_channels (int): number of input channels
        num_classes (int): number of output classes (including background)
    """

    def __init__(self, in_channels: int, num_classes: int):
        super().__init__()
        self.cls_score = nn.Linear(in_channels, num_classes)
        self.bbox_pred = nn.Linear(in_channels, num_classes * 4)

    def forward(self, inputs: Tensor) -> Tuple[Tensor, Tensor]:
        """Takes the input Tensor and pass through two separate single layer fully connected layers
            to obtain the classification logits and box regression deltas respectively.

        Args:
            inputs (Tensor): Input Tensor

        Returns:
            Tuple[Tensor, Tensor]: The classification logits and box regression deltas
        """
        if inputs.dim() == 4:
            assert list(inputs.shape[2:]) == [1, 1]
        inputs = inputs.flatten(start_dim=1)
        scores = self.cls_score(inputs)
        bbox_deltas = self.bbox_pred(inputs)

        return scores, bbox_deltas



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# MIT License

# Copyright (c) 2020 Haotian Liu and Rafael A. Rivera-Soto

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""A collection of utility functions used by YolactEdge
Modifications include:
- Removed unused python scripts in the original utils folder
- Removed unused utility functions from the original repository
- Merged utility functions from the layers folder
- Refactored config file parsing
- Modified docstrings
"""

from typing import Any, Callable, Tuple, List

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


class FastBaseTransform(torch.nn.Module):
    """
    Transform that does all operations on the GPU for improved speed.
    Maintain this as necessary.
    """

    def __init__(self, input_size: Tuple[int, int]) -> None:
        super().__init__()
        self.input_size = input_size
        # The tensor values are in BGR and are the means and standard deviation
        # values for ImageNet respectively
        if torch.cuda.is_available():
            self.mean = (
                Tensor((103.94, 116.78, 123.68)).float().cuda()[None, :, None, None]
            )
            self.std = Tensor((57.38, 57.12, 58.40)).float().cuda()[None, :, None, None]
        else:
            self.mean = Tensor((103.94, 116.78, 123.68)).float()[None, :, None, None]
            self.std = Tensor((57.38, 57.12, 58.40)).float()[None, :, None, None]

    def forward(self, img: Tensor) -> Tensor:
        """
        Args:
            img (Tensor): input image of shape (N, H, W, C)

        Returns:
            img (Tensor): transformed image of shape (N, C, H, W) where
                H == W
        """
        self.mean = self.mean.to(img.device)
        self.std = self.std.to(img.device)

        img = img.permute(0, 3, 1, 2).contiguous()
        img = F.interpolate(img, self.input_size, mode="bilinear", align_corners=False)

        img = (img - self.mean) / self.std
        img = img[:, (2, 1, 0), :, :].contiguous()
        return img


class InterpolateModule(nn.Module):
    """
    This is a module version of F.interpolate.
    Any arguments you give it just get passed along for the ride.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()
        self.args = args
        self.kwargs = kwargs

    def forward(self, inputs: Tensor) -> Tensor:
        """
        Args:
            inputs: A tensor of shape (N, C, H, W)

        Returns:
            A tensor of shape (N, C, H', W')
        """
        return F.interpolate(inputs, *self.args, **self.kwargs)


def make_net(
    in_channels: int,
    conf: List[Any],
    include_last_relu: bool = True,
) -> Tuple[nn.Sequential, int]:
    """
    A helper function to take a config setting and turn it into a network.
    Used by protonet and extrahead. Returns (network, out_channels)

    Args:
        in_channels (int): number of channels in the input
        conf (List[Any]): a list of configs to create the network
    """

    def make_layer(layer_config: Tuple[int, int, dict]) -> List[Callable]:
        nonlocal in_channels
        num_channels = layer_config[0]
        kernel_size = layer_config[1]
        if kernel_size > 0:
            layer = nn.Conv2d(in_channels, num_channels, kernel_size, **layer_config[2])
        else:
            if num_channels is None:
                layer = InterpolateModule(
                    scale_factor=-kernel_size,
                    mode="bilinear",
                    align_corners=False,
                    **layer_config[2],
                )

        in_channels = num_channels if num_channels is not None else in_channels
        return [layer, nn.ReLU(inplace=True)]

    net: List[Any] = sum([make_layer(x) for x in conf], [])
    if not include_last_relu:
        net = net[:-1]

    return nn.Sequential(*(net)), in_channels


def make_extra(num_layers: int, out_channels: int) -> Callable:
    """
    Lambda function that creates an array of num_layers alternating conv-relu if
    the num_layers is not 0.

    Args:
        num_layers (int): number of layers to create
        out_channels (int): number of channels in the output

    Returns:
        An array of alternating convolutional ReLU layers if there is at least
        one layer.
    """
    if num_layers == 0:
        return lambda x: x
    return nn.Sequential(
        *sum(
            [
                [
                    nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
                    nn.ReLU(inplace=True),
                ]
                for _ in range(num_layers)
            ],
            [],
        )
    )


def jaccard(box_a: Tensor, box_b: Tensor, is_crowd: bool = False) -> Tensor:
    """Compute the jaccard overlap of two sets of boxes.  The jaccard overlap
    is simply the intersection over union of two boxes.  Here we operate on
    ground truth boxes and default boxes. If iscrowd=True, put the crowd in box_b.
    E.g.:
        A ∩ B / A ∪ B = A ∩ B / (area(A) + area(B) - A ∩ B)

    Args:
        box_a (Tensor): Ground truth bounding boxes, Shape: [num_objects,4]
        box_b (Tensor): Prior boxes from priorbox layers, Shape: [num_priors,4]

    Returns:
        out (Tensor): Shape: [box_a.size(0), box_b.size(0)]
    """
    use_batch = True
    if box_a.dim() == 2:
        use_batch = False
        box_a = box_a[None, ...]
        box_b = box_b[None, ...]
    inter = intersect(box_a, box_b)
    area_a = (
        ((box_a[:, :, 2] - box_a[:, :, 0]) * (box_a[:, :, 3] - box_a[:, :, 1]))
        .unsqueeze(2)
        .expand_as(inter)
    )
    area_b = (
        ((box_b[:, :, 2] - box_b[:, :, 0]) * (box_b[:, :, 3] - box_b[:, :, 1]))
        .unsqueeze(1)
        .expand_as(inter)
    )
    union = area_a + area_b - inter
    out = inter / area_a if is_crowd else inter / union
    return out if use_batch else out.squeeze(0)


def point_form(boxes: Tensor) -> Tensor:
    """Convert prior_boxes to (xmin, ymin, xmax, ymax)
    representation for comparison to point form ground truth data.

    Args:
        boxes: (Tensor) center-size default boxes from priorbox layers.

    Returns:
        (Tensor) Converted xmin, ymin, xmax, ymax form of boxes.
    """
    return torch.cat(
        (
            boxes[:, :2] - boxes[:, 2:] / 2,  # xmin, ymin
            boxes[:, :2] + boxes[:, 2:] / 2,
        ),
        1,
    )  # xmax, ymax


def intersect(box_a: Tensor, box_b: Tensor) -> Tensor:
    """Resize both tensors to [A,B,2] without new malloc:
    [A,2] -> [A,1,2] -> [A,B,2]
    [B,2] -> [1,B,2] -> [A,B,2]
    Then we compute the area of intersect between box_a and box_b.

    Args:
        box_a: (Tensor) bounding boxes, Shape: [n,A,4].
        box_b: (Tensor) bounding boxes, Shape: [n,B,4].

    Returns:
        (Tensor) intersection area, Shape: [n,A,B].
    """
    num = box_a.size(0)
    set_a = box_a.size(1)
    set_b = box_b.size(1)

    max_xy = torch.min(
        box_a[:, :, 2:].unsqueeze(2).expand(num, set_a, set_b, 2),
        box_b[:, :, 2:].unsqueeze(1).expand(num, set_a, set_b, 2),
    )
    min_xy = torch.max(
        box_a[:, :, :2].unsqueeze(2).expand(num, set_a, set_b, 2),
        box_b[:, :, :2].unsqueeze(1).expand(num, set_a, set_b, 2),
    )
    inter = torch.clamp((max_xy - min_xy), min=0)
    return inter[:, :, :, 0] * inter[:, :, :, 1]


def decode(loc: Tensor, priors: Tensor, use_yolo_regressors: bool = False) -> Tensor:
    """
    Decode predicted bbox coordinates using the same scheme
    employed by Yolov2: https://arxiv.org/pdf/1612.08242.pdf
        b_x = (sigmoid(pred_x) - .5) / conv_w + prior_x
        b_y = (sigmoid(pred_y) - .5) / conv_h + prior_y
        b_w = prior_w * exp(loc_w)
        b_h = prior_h * exp(loc_h)

    Note that loc is inputed as [(s(x)-.5)/conv_w, (s(y)-.5)/conv_h, w, h]
    while priors are inputed as [x, y, w, h] where each coordinate
    is relative to size of the image (even sigmoid(x)). We do this
    in the network by dividing by the 'cell size', which is just
    the size of the convouts.

    Also note that prior_x and prior_y are center coordinates which
    is why we have to subtract .5 from sigmoid(pred_x and pred_y).

    Args:
        loc (Tensor): The predicted bounding boxes of size [num_priors, 4]
        priors (Tensor): The priorbox coords with size [num_priors, 4]
        use_yolo_regressors (bool): Whether or not to use the YOLO regressors.

    Returns:
        boxes (Tensor): A tensor of decoded relative coordinates in point form
            form with size [num_priors, 4]
    """
    if use_yolo_regressors:
        boxes = torch.cat(
            (loc[:, :2] + priors[:, :2], priors[:, 2:] * torch.exp(loc[:, 2:])), 1
        )
        boxes = point_form(boxes)
    else:
        variances = [0.1, 0.2]
        boxes = torch.cat(
            (
                priors[:, :2] + loc[:, :2] * variances[0] * priors[:, 2:],
                priors[:, 2:] * torch.exp(loc[:, 2:] * variances[1]),
            ),
            1,
        )
        boxes[:, :2] -= boxes[:, 2:] / 2
        boxes[:, 2:] += boxes[:, :2]

    return boxes


def sanitize_coordinates(
    _x1: Tensor, _x2: Tensor, img_size: int, padding: int = 0, cast: bool = True
) -> Tuple[Tensor, Tensor]:
    """Sanitizes the input coordinates so that x1 < x2, x1 != x2, x1 >= 0,
    and x2 <= image_size. Also converts from relative to absolute coordinates and
    casts the results to long tensors. If cast is false, the result won't be cast
    to longs.

    Args:
        _x1 (Tensor): input coordinate of x1.
        _x2 (Tensor): input coordinate of x2.
        img_size (int): the dimensions of the image.
        padding (int, optional): Padding value. Defaults to 0.
        cast (bool, optional): Cast the results to long tensors. Defaults to True.

    Returns:
        x_1, x_2: The sanitized input coordinates of _x1 and _x2 respectively.
    """
    _x1 = _x1 * img_size
    _x2 = _x2 * img_size
    if cast:
        _x1 = _x1.long()
        _x2 = _x2.long()
    x_1 = torch.min(_x1, _x2)
    x_2 = torch.max(_x1, _x2)
    x_1 = torch.clamp(x_1 - padding, min=0)
    x_2 = torch.clamp(x_2 + padding, max=img_size)
    return x_1, x_2


def crop(  # pylint: disable=too-many-locals
    masks: Tensor, boxes: Tensor, padding: int = 1
) -> Tensor:
    """ "Crop" predicted masks by zeroing out everything not in the predicted bbox.
    Vectorized by Chong Zhou.

    Args:
        masks (Tensor): Uncropped mask tensor values in uint8
        boxes (Tensor): x1, y1, x2, y2 values of the bounding box detection
        padding (int, optional): Padding value. Defaults to 1.

    Returns:
        out (Tensor): cropped mask tensor
    """
    height, width, batch = masks.size()
    x_1, x_2 = sanitize_coordinates(
        boxes[:, 0], boxes[:, 2], width, padding, cast=False
    )
    y_1, y_2 = sanitize_coordinates(
        boxes[:, 1], boxes[:, 3], height, padding, cast=False
    )

    rows = (
        torch.arange(width, device=masks.device, dtype=x_1.dtype)
        .view(1, -1, 1)
        .expand(height, width, batch)
    )

    cols = (
        torch.arange(height, device=masks.device, dtype=x_1.dtype)
        .view(-1, 1, 1)
        .expand(height, width, batch)
    )

    masks_left = rows >= x_1.view(1, 1, -1)
    masks_right = rows < x_2.view(1, 1, -1)
    masks_up = cols >= y_1.view(1, 1, -1)
    masks_down = cols < y_2.view(1, 1, -1)
    crop_mask = masks_left * masks_right * masks_up * masks_down
    out = masks * crop_mask.float()
    return out
